#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
#include <io.h>
#include <dos.h>
#include <sys\stat.h>
#include <sys\types.h>
#include <string.h>

#define MAXITEMS 4096

typedef struct
{
	char name[13];
	unsigned long size, offset;
} Item_T;

long numitems, indexoffset, indexsize;
long mergenumitems, mergeoffset, mergesize;
Item_T item[MAXITEMS], mergeitem[MAXITEMS];

static int mergeditem[MAXITEMS];
static char filespec[MAXITEMS][128], filelist[MAXITEMS][16];
static long fileleng[MAXITEMS];
char header[4], buf[65536];
char *itembuf[MAXITEMS];
long dathandle;

void findfiles(char *dafilespec)
{
	struct find_t fileinfo;
	long i, j, fil, daspeclen;
	char daspec[128];

	strcpy(daspec,dafilespec);
	for(daspeclen=0;daspec[daspeclen]!=0;daspeclen++);
	while ((daspec[daspeclen] != '\\') && (daspeclen > 0)) daspeclen--;
	if (daspeclen > 0) daspec[daspeclen++] = '\\';

	if (_dos_findfirst(dafilespec,_A_NORMAL,&fileinfo) != 0) return;
	do
	{
		strcpy(&filelist[numitems][0],fileinfo.name);
		fileleng[numitems] = fileinfo.size;
		filelist[numitems][12] = (char)(fileleng[numitems]&255);
		filelist[numitems][13] = (char)((fileleng[numitems]>>8)&255);
		filelist[numitems][14] = (char)((fileleng[numitems]>>16)&255);
		filelist[numitems][15] = (char)((fileleng[numitems]>>24)&255);

		strcpy(&filespec[numitems][0],daspec);
		strcpy(&filespec[numitems][daspeclen],fileinfo.name);
		strcpy(&item[numitems].name,fileinfo.name);

		item[numitems].size = fileleng[numitems];
		if ((fil = open(filespec[numitems],O_BINARY|O_RDONLY,S_IREAD)) == -1)
		{
			printf("Error: %s not found\n",filespec[i]);
			exit(0);
		}
		itembuf[numitems] = (char *)malloc(fileleng[numitems]);
		if (itembuf[numitems] == NULL)
		{
			printf("Not enough memory for item\n");
			exit(-1);
		}
		read(fil,itembuf[numitems],fileleng[numitems]);
		close(fil);
		numitems++;
		if (numitems > MAXITEMS)
		{
			printf("FATAL ERROR: TOO MANY FILES SELECTED! (MAX is 4096)\n");
			exit(0);
		}		
	} while (_dos_findnext(&fileinfo) == 0);
}

int main(int argc, char *argv[])
{
	long i, j, k, l, fil2;
	FILE *fil;
	char buffer[80], tmp[256], tmp2[256];
	
	if (argc < 3)
	{
		printf("COMPRESS v1.66	- By Micheal Muniko\n");
		printf("Syntax: compress [data file] [script file]\n");
		exit(0);
	}
	fil = fopen(argv[2], "r");
	if (fil != NULL)
	{
		while (fgets(buffer, 80, fil) != NULL)
		{
			sscanf(buffer, "%s", tmp);
			if (strstr("@link", tmp) != 0)
			{
				sscanf(buffer, "%s %s", tmp, tmp2);
				if ((dathandle = open(tmp2, O_BINARY|O_RDONLY,S_IREAD)) != -1)
				{
					read(dathandle, &header, 4);
					read(dathandle, &mergenumitems, 2);
					read(dathandle, &mergeoffset, 4);
					read(dathandle, &mergesize, 4);
					lseek(dathandle, mergeoffset, SEEK_SET);
					read(dathandle, &mergeitem[0], mergesize);
					for(i=0;i<mergenumitems;i++)
					{
						item[numitems] = mergeitem[i];
						itembuf[numitems] = (char *)malloc(item[numitems].size);
						lseek(dathandle, mergeitem[i].offset, SEEK_SET);
						read(dathandle, itembuf[numitems], mergeitem[i].size);
						mergeditem[numitems] = 1;
						numitems++;
					}
					close(dathandle);
				}
			}		
			else
				findfiles(buffer);
		}
	}
	else {
		printf("Cannot open %s\n", argv[2]);
		exit(1);
	}
	fclose(fil);
	if ((dathandle = open(argv[1], O_BINARY|O_WRONLY|O_CREAT|O_TRUNC,S_IWRITE)) == -1)
	{
		printf("Cannot open %s\n", argv[1]);
		exit(1);
	}
	header[0] = 'D';
	header[1] = 'A';
	header[2] = 'T';
	header[3] = 0x1a;
	write(dathandle,&header,4);
	write(dathandle,&numitems,2);
	indexoffset = tell(dathandle)+8;
	for(i=0;i<numitems;i++)
	{
		indexoffset += item[i].size;
	}
	indexsize = sizeof(Item_T)*numitems;
	write(dathandle,&indexoffset,4);
	write(dathandle,&indexsize,4);
	for(i=0;i<numitems;i++)
	{
		printf("Adding item %s...\n",item[i].name);
		item[i].offset = tell(dathandle);
		if (write(dathandle, itembuf[i], item[i].size) < item[i].size)
		{
			close(dathandle);
			printf("OUT OF HD SPACE!\n");
			exit(0);
		}
	}
	write(dathandle,&item[0],sizeof(Item_T)*numitems);
	close(dathandle);
	for(i=0;i<numitems;i++)
	{
		if (itembuf[i] != NULL)
			free(itembuf[i]);
	}
	printf("DAT file created\n");
	return(0);
}
